from __future__ import annotations

from typing import Any

from bailo.core.client import Client


class AccessRequest:
    """Represent a review within Bailo.

    A review can either be access to a model or to a specific release.

    :param client: A client object that is used to make requests to bailo
    :param name: The name of the access request
    :param model_id: The unique model id of the model that the access request is being made with
    :param schema_id: An ID for the schema within Bailo
    :param metadata: A metadata object
    :param access_request_id: The Unique ID for this access request
    :param deleted: Whether the access request has been deleted
    """

    def __init__(
        self,
        client: Client,
        model_id: str,
        schema_id: str,
        metadata: Any,
        access_request_id: str,
        created_by: str,
        deleted: bool = False,
    ) -> None:
        self.client = client
        self.model_id = model_id
        self.schema_id = schema_id
        self.metadata = metadata
        self.access_request_id = access_request_id
        self.created_by = created_by
        self.deleted = deleted

    @classmethod
    def from_id(cls, client: Client, model_id: str, access_request_id: str) -> AccessRequest:
        """Return an existing review from Bailo given it's unique ID.

        :param client: A client object used to interact with Bailo
        :param model_id: A unique model ID within Bailo
        :param access_request_id: A unique ID for an access request
        """
        json_access_request = client.get_access_request(model_id, access_request_id)["accessRequest"]

        deleted = json_access_request["deleted"]
        created_by = json_access_request["createdBy"]
        metadata = json_access_request["metadata"]

        schema_id = json_access_request["schemaId"]

        return cls(
            client,
            model_id,
            schema_id,
            metadata,
            access_request_id,
            created_by,
            deleted,
        )

    @classmethod
    def create(cls, client: Client, model_id: str, schema_id: str, metadata: Any) -> AccessRequest:
        """Make an access request for the model.

        Posts an access request to Bailo to be reviewed

        :param client: A client object used to interact with Bailo
        :param name: The name of the access request
        :param model_id: A unique model ID within Bailo
        :param schema_id: A unique schema ID
        :return: JSON response object
        """
        access_request_json = client.post_access_request(model_id, metadata, schema_id)["accessRequest"]

        deleted = access_request_json["deleted"]
        access_request_id = access_request_json["id"]
        metadata = access_request_json["metadata"]
        created_by = access_request_json["createdBy"]

        return cls(
            client,
            model_id,
            schema_id,
            metadata,
            access_request_id,
            created_by,
            deleted,
        )

    def delete(self) -> bool:
        """Delete the access request on Bailo.

        :return: A message confirming the removal of the access request.
        """
        self.client.delete_access_request(self.model_id, self.access_request_id)
        return True

    def update(self):
        """Update the current state of the access request to Bailo."""
        self.client.patch_access_request(self.model_id, self.access_request_id, metadata=self.metadata)

    def __str__(self) -> str:
        return f"Access Request: {self.metadata['overview']['name']} - {self.model_id}"

    def __repr__(self):
        return f"{self.__class__.__name__}({self.model_id},{self.schema_id})"

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.access_request_id == other.access_request_id



from __future__ import annotations

from typing import Any

from bailo.core.client import Client
from bailo.core.enums import ModelVisibility
from bailo.helper.release import Release
from bailo.core.exceptions import BailoException
from bailo.core.utils import NestedDict

import os
import shutil
import tempfile
from semantic_version import Version

try:
    import mlflow

    ml_flow = True
except ImportError:
    ml_flow = False


class Model:
    """Represent a model within Bailo.

    :param client: A client object used to interact with Bailo
    :param model_id: A unique ID for the model
    :param name: Name of model
    :param description: Description of model
    :param visibility: Visibility of model, using ModelVisibility enum (e.g Public or Private), defaults to None
    """

    def __init__(
        self,
        client: Client,
        model_id: str,
        name: str,
        description: str,
        visibility: ModelVisibility | None = None,
    ) -> None:
        self.client = client

        self.model_id = model_id
        self.name = name
        self.description = description
        self.visibility = visibility

        self.model_card = None
        self.model_card_version = None
        self.model_card_schema = None

    @classmethod
    def create(
        cls,
        client: Client,
        name: str,
        description: str,
        team_id: str,
        visibility: ModelVisibility | None = None,
    ) -> Model:
        """Build a model from Bailo and uploads it.

        :param client: A client object used to interact with Bailo
        :param name: Name of model
        :param description: Description of model
        :param team_id: A unique team ID
        :param visibility: Visibility of model, using ModelVisibility enum (e.g Public or Private), defaults to None
        :return: Model object
        """
        res = client.post_model(name=name, description=description, team_id=team_id, visibility=visibility)
        model = cls(
            client=client,
            model_id=res["model"]["id"],
            name=name,
            description=description,
            visibility=visibility,
        )

        model.__unpack(res["model"])

        return model

    @classmethod
    def from_id(cls, client: Client, model_id: str) -> Model:
        """Return an existing model from Bailo.

        :param client: A client object used to interact with Bailo
        :param model_id: A unique model ID
        :return: A model object
        """
        res = client.get_model(model_id=model_id)["model"]
        model = cls(
            client=client,
            model_id=model_id,
            name=res["name"],
            description=res["description"],
        )
        model.__unpack(res)

        model.get_card_latest()

        return model

    def update(self) -> None:
        """Upload and retrieves any changes to the model summary on Bailo."""
        res = self.client.patch_model(
            model_id=self.model_id,
            name=self.name,
            description=self.description,
            visibility=self.visibility,
        )
        self.__unpack(res["model"])

    def update_model_card(self, model_card: dict[str, Any] | None = None) -> None:
        """Upload and retrieves any changes to the model card on Bailo.

        :param model_card: Model card dictionary, defaults to None

        ..note:: If a model card is not provided, the current model card attribute value is used
        """
        if model_card is None:
            model_card = self.model_card
        res = self.client.put_model_card(model_id=self.model_id, metadata=model_card)
        self.__unpack_mc(res["card"])

    def card_from_schema(self, schema_id: str) -> None:
        """Create a model card using a schema on Bailo.

        :param schema_id: A unique schema ID
        """
        res = self.client.model_card_from_schema(model_id=self.model_id, schema_id=schema_id)
        self.__unpack_mc(res["card"])

    def card_from_model(self):
        """Copy a model card from a different model (not yet implemented).

        :raises NotImplementedError: Not implemented error
        """
        raise NotImplementedError

    def card_from_template(self):
        """Create a model card using a template (not yet implemented).

        :raises NotImplementedError: Not implemented error
        """
        raise NotImplementedError

    def get_card_latest(self) -> None:
        """Get the latest model card from Bailo."""
        res = self.client.get_model(model_id=self.model_id)
        if "card" in res["model"]:
            self.__unpack_mc(res["model"]["card"])
        else:
            raise BailoException(f"A model card doesn't exist for model {self.model_id}")

    def get_card_revision(self, version: str) -> None:
        """Get a specific model card revision from Bailo.

        :param version: Model card version
        """
        res = self.client.get_model_card(model_id=self.model_id, version=version)
        self.__unpack_mc(res["modelCard"])

    def create_experiment(
        self,
    ) -> Experiment:
        return Experiment.create(model=self)

    def create_release(
        self,
        version: Version | str,
        notes: str,
        files: list[str] | None = None,
        images: list[str] | None = None,
        minor: bool = False,
        draft: bool = True,
    ) -> Release:
        """Call the Release.create method to build a release from Bailo and upload it.

        :param version: A semantic version for the release
        :param notes: Notes on release, defaults to ""
        :param files: A list of files for release, defaults to []
        :param images: A list of images for release, defaults to []
        :param minor: Is a minor release?, defaults to False
        :param draft: Is a draft release?, defaults to True
        :return: Release object
        """
        if self.model_card_schema:
            return Release.create(
                client=self.client,
                model_id=self.model_id,
                version=version,
                notes=notes,
                model_card_version=self.model_card_version,
                files=files,
                images=images,
                minor=minor,
                draft=draft,
            )
        raise BailoException("Create a model card before creating a release")

    def get_releases(self) -> list[Release]:
        """Get all releases for the model.

        :return: List of Release objects
        """
        res = self.client.get_all_releases(model_id=self.model_id)
        releases = []

        for release in res["releases"]:
            releases.append(self.get_release(version=release["semver"]))

        return releases

    def get_release(self, version: Version | str) -> Release:
        """Call the Release.from_version method to return an existing release from Bailo.

        :param version: A semantic version for the release
        :return: Release object
        """
        return Release.from_version(self.client, self.model_id, version)

    def get_latest_release(self):
        """Get the latest release for the model from Bailo.

        :return: Release object
        """
        releases = self.get_releases()
        if releases == []:
            raise BailoException("This model has no releases.")
        return max(releases)

    def get_images(self):
        """Get all model image references for the model.

        :return: List of images
        """
        res = self.client.get_all_images(model_id=self.model_id)

        return res["images"]

    def get_image(self):
        """Get a model image reference.

        :raises NotImplementedError: Not implemented error.
        """
        raise NotImplementedError

    def get_roles(self):
        """Get all roles for the model.

        :return: List of roles
        """
        res = self.client.get_model_roles(model_id=self.model_id)

        return res["roles"]

    def get_user_roles(self):
        """Get all user roles for the model.

        :return: List of user roles
        """
        res = self.client.get_model_user_roles(model_id=self.model_id)

        return res["roles"]

    def __unpack(self, res):
        self.model_id = res["id"]
        self.name = res["name"]
        self.description = res["description"]

        if res["visibility"] == "private":
            self.visibility = ModelVisibility.PRIVATE
        else:
            self.visibility = ModelVisibility.PUBLIC

    def __unpack_mc(self, res):
        self.model_card_version = res["version"]
        self.model_card_schema = res["schemaId"]

        try:
            self.model_card = res["metadata"]
        except KeyError:
            self.model_card = None


class Experiment:
    """Represent an experiment locally.

    :param model: A Bailo model object which the experiment is being run on
    :param raw: Raw information about the experiment runs

    .. code-block:: python

       experiment = model.create_experiment()
       for x in range(5):
           experiment.start_run()
           experiment.log_params({"lr": 0.01})
           ### INSERT MODEL TRAINING HERE ###
           experiment.log_metrics("accuracy": 0.86)
           experiment.log_artifacts(["weights.pth"])

       experiment.publish(mc_loc="performance.performanceMetrics", run_id=1)

    """

    def __init__(
        self,
        model: Model,
    ):
        self.model = model
        self.raw = []
        self.run = -1
        self.temp_dir = os.path.join(tempfile.gettempdir(), "bailo_runs")

    @classmethod
    def create(
        cls,
        model: Model,
    ) -> Experiment:
        """Create an experiment locally.

        :param model: A Bailo model object which the experiment is being run on
        :return: Experiment object
        """

        return cls(model=model)

    def start_run(self, is_mlflow: bool = False):
        """Starts a new experiment run.

        :param is_mlflow: Marks a run as MLFlow
        """
        self.run += 1

        self.run_data = {"run": self.run, "params": [], "metrics": [], "artifacts": [], "dataset": ""}

        self.raw.append(self.run_data)

        if not is_mlflow:
            print(f"Bailo tracking run {self.run}.")

    def log_params(self, params: dict[str, Any]):
        """Logs parameters to the current run.

        :param params: Dictionary of parameters to be logged
        """
        self.run_data["params"].append(params)

    def log_metrics(self, metrics: dict[str, Any]):
        """Logs metrics to the current run.

        :param metrics: Dictionary of metrics to be logged
        """
        self.run_data["metrics"].append(metrics)

    def log_artifacts(self, artifacts: list):
        """Logs artifacts to the current run.

        :param artifacts: A list of artifact paths to be logged
        """
        self.run_data["artifacts"].extend(artifacts)

    def log_dataset(self, dataset: str):
        """Logs a dataset to the current run.

        :param dataset: Arbitrary title of dataset
        """
        self.run_data["dataset"] = dataset

    def from_mlflow(self, tracking_uri: str, experiment_id: str):
        """Imports information from an MLFlow Tracking experiment.

        :param tracking_uri: MLFlow Tracking server URI
        :param experiment_id: MLFlow Tracking experiment ID
        :raises ImportError: Import error if MLFlow not installed
        """
        if ml_flow:
            client = mlflow.tracking.MlflowClient(tracking_uri=tracking_uri)
            runs = client.search_runs(experiment_id)

            for run in runs:
                data = run.data
                info = run.info
                inputs = run.inputs

                artifact_uri = info.artifact_uri
                run_id = info.run_id
                status = info.status
                datasets = inputs.dataset_inputs
                datasets_str = [dataset.name for dataset in datasets]

                artifacts = []

                # MLFlow run must be status FINISHED
                if status != "FINISHED":
                    continue

                if len(mlflow.artifacts.list_artifacts(artifact_uri=artifact_uri)) > 0:
                    mlflow_dir = os.path.join(self.temp_dir, f"mlflow_{run_id}")
                    mlflow.artifacts.download_artifacts(artifact_uri=artifact_uri, dst_path=mlflow_dir)
                    artifacts.append(mlflow_dir)

                self.start_run(is_mlflow=True)
                self.log_params(data.params)
                self.log_metrics(data.metrics)
                self.log_artifacts(artifacts)
                self.log_dataset("".join(datasets_str))
                self.run_data["run"] = info.run_id
        else:
            raise ImportError("Optional MLFlow dependencies (needed for this method) are not installed.")

    def publish(self, mc_loc: str, run_id: str, semver: str = "0.1.0", notes: str = ""):
        """Publishes a given experiments results to the model card.

        :param mc_loc: Location of metrics in the model card (e.g. performance.performanceMetrics)
        :param run_id: Local experiment run ID to be selected
        :param semver: Semantic version of release to create (if artifacts present), defaults to 0.1.0 or next
        :param notes: Notes for release, defaults to ""

        ..note:: mc_loc is dependent on the model card schema being used
        """
        mc = self.model.model_card
        if mc is None:
            raise BailoException("Model card needs to be populated before publishing an experiment.")

        mc = NestedDict(mc)

        if len(self.raw) > 0:
            for run in self.raw:
                if run["run"] == run_id:
                    sel_run = run
                    break
            else:
                raise NameError(f"Run {run_id} does not exist.")
        else:
            raise BailoException(f"This experiment has no runs to publish.")

        values = []

        for metric in sel_run["metrics"]:
            for k, v in metric.items():
                values.append({"name": k, "value": v})

        # Updating the model card
        parsed_values = [{"dataset": sel_run["dataset"], "datasetMetrics": values}]
        mc[tuple(mc_loc.split("."))] = parsed_values
        self.model.update_model_card(model_card=mc)

        # Creating a release and uploading artifacts (if artifacts present)
        artifacts = sel_run["artifacts"]
        if len(artifacts) > 0:
            # Create new release
            try:
                release_latest_version = self.model.get_latest_release().version
                release_new_version = release_latest_version.next_minor()
            except:
                release_new_version = semver

            run_id = sel_run["run"]
            notes = f"{notes} (Run ID: {run_id})"
            release_new = self.model.create_release(version=release_new_version, minor=True, notes=notes)

            for artifact in artifacts:
                release_new.upload(path=artifact)

            if os.path.exists(self.temp_dir) and os.path.isdir(self.temp_dir):
                shutil.rmtree(self.temp_dir)



from __future__ import annotations

import os
import shutil
from io import BytesIO
from typing import Any

from bailo.core.client import Client
from semantic_version import Version


class Release:
    def __init__(
        self,
        client: Client,
        model_id: str,
        version: Version | str,
        model_card_version: int | None = None,
        notes: str = "",
        files: list[str] | None = None,
        images: list[str] | None = None,
        minor: bool = False,
        draft: bool = True,
    ) -> None:
        """Represent a release within Bailo.

        :param client: A client object used to interact with Bailo
        :param model_id: A unique model ID
        :param version: A semantic version for the release
        :param model_card_version: Version of the model card
        :param notes: Notes on release
        :param files: (optional) A list of files for release
        :param images: (optional) A list of images for release
        :param minor: Is a minor release?
        :param draft: Is a draft release?

        ..note:: Currently files and images are stored as string references
        """
        self.client = client
        self.model_id = model_id

        if files is None:
            files = []

        if images is None:
            images = []

        if isinstance(version, str):
            version = Version(version)
        self.version = version

        self.model_card_version = model_card_version
        self.minor = minor
        self.notes = notes
        self.files = files
        self.images = images
        self.draft = draft
        self.files = files

    @classmethod
    def create(
        cls,
        client: Client,
        model_id: str,
        version: Version | str,
        notes: str,
        model_card_version: int | None = None,
        files: list[str] | None = None,
        images: list[str] | None = None,
        minor: bool = False,
        draft: bool = True,
    ) -> Release:
        """Build a release from Bailo and uploads it.

        :param client: A client object used to interact with Bailo
        :param model_id: A Unique Model ID
        :param version: A semantic version of a model release
        """
        if files is None:
            files = []

        if images is None:
            images = []

        client.post_release(
            model_id,
            str(version),
            notes,
            files,
            images,
            model_card_version,
            minor,
            draft,
        )

        return cls(
            client,
            model_id,
            version,
            model_card_version,
            notes,
            files,
            images,
            minor,
            draft,
        )

    @classmethod
    def from_version(cls, client: Client, model_id: str, version: Version | str) -> Release:
        """Return an existing release from Bailo.

        :param client: A client object used to interact with Bailo
        :param model_id: A Unique Model ID
        :param version: A semantic version of a model release
        """
        res = client.get_release(model_id, str(version))["release"]

        model_card_version = res["modelCardVersion"]
        notes = res["notes"]
        files = res["fileIds"]
        images = res["images"]
        minor = res["minor"]
        draft = res["draft"]

        return cls(
            client,
            model_id,
            version,
            model_card_version,
            notes,
            files,
            images,
            minor,
            draft,
        )

    def download(self, filename: str, write: bool = True, path: str | None = None) -> Any:
        """Give returns a Reading object given the file id.

        :param filename: The name of the file to retrieve
        :param write: Bool to determine if writing file to disk, defaults to True
        :param path: Local path to write file to (if write set to True)

        :return: A JSON response object
        """
        res = self.client.get_download_by_filename(self.model_id, str(self.version), filename)

        if write:
            if path is None:
                path = filename
            with open(path, "wb") as f:
                f.write(res.content)

        return res

    def upload(self, path: str, data: BytesIO | None = None) -> str:
        """Upload a file to the release.

        :param path: The path, or name of file or directory to be uploaded
        :param data: A BytesIO object if not loading from disk

        :return: The unique file ID of the file uploaded
        ..note:: If path provided is a directory, it will be uploaded as a zip
        """
        name = os.path.split(path)[1]

        if data is None:
            if is_zip := os.path.isdir(path):
                shutil.make_archive(name, "zip", path)
                path = f"{name}.zip"
                name = path

            with open(path, "rb") as f:
                res = self.client.simple_upload(self.model_id, name, f).json()

            if is_zip:
                os.remove(path)
        else:
            res = self.client.simple_upload(self.model_id, name, data).json()

        self.files.append(res["file"]["id"])
        self.update()
        return res["file"]["id"]

    def update(self) -> Any:
        """Update the any changes to this release on Bailo.

        :return: JSON Response object
        """
        return self.client.put_release(
            self.model_id,
            str(self.version),
            self.notes,
            self.draft,
            self.files,
            self.images,
        )

    def delete(self) -> Any:
        """Delete a release from Bailo.

        :return: JSON Response object
        """
        self.client.delete_release(self.model_id, str(self.version))
        return True

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({str(self)})"

    def __str__(self) -> str:
        return f"{self.model_id} v{self.version}"

    def __eq__(self, other) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.version == other.version

    def __ne__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self != other

    def __lt__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.version < other.version

    def __le__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.version <= other.version

    def __gt__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.version > other.version

    def __ge__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.version >= other.version

    def __hash__(self) -> int:
        return hash((self.model_id, self.version))



from __future__ import annotations


class BailoException(Exception):
    """General exception for Bailo response errors."""


class ResponseException(Exception):
    """Exception used if an endpoint gave no response."""



from __future__ import annotations

from enum import Enum


class ValuedEnum(str, Enum):
    def __str__(self) -> str:
        return str(self.value)


class ModelVisibility(ValuedEnum):
    """Whether a model is publicly visible or not."""

    PRIVATE = "private"
    PUBLIC = "public"


class SchemaKind(ValuedEnum):
    """A type of schema."""

    MODEL = "model"
    ACCESS_REQUEST = "accessRequest"


class Role(ValuedEnum):
    """A reviewing role."""

    OWNER = "owner"
    MODEL_TECHNICAL_REVIEWER = "mtr"
    MODEL_SENIOR_RESPONSIBLE_OFFICER = "msro"



from __future__ import annotations

import shutil
from io import BytesIO
from typing import Any

from bailo.core.agent import Agent, TokenAgent
from bailo.core.enums import ModelVisibility, SchemaKind
from bailo.core.utils import filter_none


class Client:
    """Create a Client object that can be used to talk to the website.

    :param url: Url of bailo website
    :param agent: An agent object to handle requests
    """

    def __init__(self, url: str, agent: Agent = Agent()):
        self.url = url.rstrip("/") + "/api"
        self.agent = agent

    def post_model(
        self,
        name: str,
        description: str,
        team_id: str,
        visibility: ModelVisibility | None = None,
    ):
        """Create a model.

        :param name: Name of the model
        :param description: Description of the model
        :param visibility: Enum to define model visibility (e.g public or private)
        :return: JSON response object
        """
        if visibility is not None:
            visibility = str(visibility)

        filtered_json = filter_none(
            {
                "name": name,
                "description": description,
                "visibility": visibility,
                "teamId": team_id,
            }
        )

        return self.agent.post(
            f"{self.url}/v2/models",
            json=filtered_json,
        ).json()

    def get_models(
        self,
        task: str | None = None,
        libraries: list[str] | None = None,
        filters: list[str] | None = None,
        search: str = "",
    ):
        """Find and returns a list of models based on provided search terms.

        :param task: Model task (e.g. image classification), defaults to None
        :param libraries: Model library (e.g. TensorFlow), defaults to []
        :param filters: Custom filters, defaults to []
        :param search: String to be located in model cards, defaults to ""
        :return: JSON response object
        """
        if libraries is None:
            libraries = []

        if filters is None:
            filters = []

        return self.agent.get(
            f"{self.url}/v2/models/search",
            params={
                "task": task,
                "libraries": libraries,
                "filters": filters,
                "search": search,
            },
        ).json()

    def get_model(
        self,
        model_id: str,
    ):
        """Retrieve a specific model using its unique ID.

        :param model_id: Unique model ID
        :return: JSON response object
        """
        return self.agent.get(
            f"{self.url}/v2/model/{model_id}",
        ).json()

    def patch_model(
        self,
        model_id: str,
        name: str | None = None,
        description: str | None = None,
        visibility: str | None = None,
    ):
        """Update a specific model using its unique ID.

        :param model_id: Unique model ID
        :param name: Name of the model, defaults to None
        :param description: Description of the model, defaults to None
        :param visibility: Enum to define model visibility (e.g. public or private), defaults to None
        :return: JSON response object
        """
        filtered_json = filter_none({"name": name, "description": description, "visibility": visibility})

        return self.agent.patch(f"{self.url}/v2/model/{model_id}", json=filtered_json).json()

    def get_model_card(
        self,
        model_id: str,
        version: str,
    ):
        """Retrieve a specific model card, using the unique model ID and version.

        :param model_id: Unique model ID
        :param version: Model card version
        :return: JSON response object
        """
        return self.agent.get(
            f"{self.url}/v2/model/{model_id}/model-card/{version}",
        ).json()

    def put_model_card(
        self,
        model_id: str,
        metadata: Any,
    ):
        """Update the latest model card, using the unique model ID.

        :param model_id: Unique model ID
        :param metadata: Metadata object, defined by model card schema
        :return: JSON response object
        """
        return self.agent.put(
            f"{self.url}/v2/model/{model_id}/model-cards",
            json={
                "metadata": metadata,
            },
        ).json()

    def model_card_from_schema(
        self,
        model_id: str,
        schema_id: str,
    ):
        """
        Create a model card using a given schema ID.

        :param model_id: Unique model ID
        :param schema_id: Unique model card schema ID
        :return: JSON response object
        """
        return self.agent.post(
            f"{self.url}/v2/model/{model_id}/setup/from-schema",
            json={
                "schemaId": schema_id,
            },
        ).json()

    def post_release(
        self,
        model_id: str,
        release_version: str,
        notes: str,
        file_ids: list[str],
        images: list[str],
        model_card_version: int | None = None,
        minor: bool | None = False,
        draft: bool | None = False,
    ):
        """
        Create a new model release.

        :param model_id: Unique model ID
        :param model_card_version: Model card version
        :param release_version: Release version
        :param notes: Notes on release
        :param file_ids: Files for release
        :param images: Images for release
        :param minor: Signifies a minor release, defaults to False
        :param draft: Signifies a draft release, defaults to False
        :return: JSON response object
        """
        filtered_json = filter_none(
            {
                "modelCardVersion": model_card_version,
                "semver": release_version,
                "notes": notes,
                "minor": minor,
                "draft": draft,
                "fileIds": file_ids,
                "images": images,
            }
        )
        return self.agent.post(f"{self.url}/v2/model/{model_id}/releases", json=filtered_json).json()

    def put_release(
        self,
        model_id: str,
        release_version: str,
        notes: str,
        draft: bool,
        file_ids: list[str],
        images: list[str],
    ):
        """
        Create a new model release.

        :param model_id: Unique model ID
        :param model_card_version: Model card version
        :param release_version: Release version
        :param notes: Notes on release
        :param file_ids: Files for release
        :param images: Images for release
        :param minor: Signifies a minor release, defaults to False
        :param draft: Signifies a draft release, defaults to False
        :return: JSON response object
        """
        return self.agent.put(
            f"{self.url}/v2/model/{model_id}/release/{release_version}",
            json={
                "notes": notes,
                "draft": draft,
                "fileIds": file_ids,
                "images": images,
            },
        ).json()

    def get_all_releases(
        self,
        model_id: str,
    ):
        """
        Get all releases for a model.

        :param model_id: Unique model ID
        :return: JSON response object
        """
        return self.agent.get(
            f"{self.url}/v2/model/{model_id}/releases",
        ).json()

    def get_release(self, model_id: str, release_version: str):
        """
        Get a specific model release.

        :param model_id: Unique model ID
        :param release_version: Release version
        :return: JSON response object
        """
        return self.agent.get(
            f"{self.url}/v2/model/{model_id}/release/{release_version}",
        ).json()

    def delete_release(
        self,
        model_id: str,
        release_version: str,
    ):
        """
        Delete a specific model release.

        :param model_id: Unique model ID
        :param release_version: Release version
        :return: JSON response object
        """
        return self.agent.delete(
            f"{self.url}/v2/model/{model_id}/release/{release_version}",
        ).json()

    def get_files(
        self,
        model_id: str,
    ):
        """
        Get files for a model.

        :param model_id: Unique model ID
        :return: JSON response object
        """
        return self.agent.get(
            f"{self.url}/v2/model/{model_id}/files",
        ).json()

    def get_download_file(
        self,
        model_id: str,
        file_id: str,
    ):
        """Download a specific file by it's id.

        :param model_id: Unique model ID
        :param file_id: Unique file ID
        :return: The unique file ID
        """
        if isinstance(self.agent, TokenAgent):
            return self.agent.get(
                f"{self.url}/v2/token/model/{model_id}/file/{file_id}/download", stream=True, timeout=10_000
            )
        else:
            return self.agent.get(
                f"{self.url}/v2/model/{model_id}/file/{file_id}/download", stream=True, timeout=10_000
            )

    def get_download_by_filename(
        self,
        model_id: str,
        semver: str,
        filename: str,
    ):
        """Download a specific file.

        :param model_id: Unique model ID
        :param semver: Semver of the release
        :param filename: The filename trying to download from
        :return: The filename
        """
        if isinstance(self.agent, TokenAgent):
            return self.agent.get(
                f"{self.url}/v2/token/model/{model_id}/release/{semver}/file/{filename}/download",
                stream=True,
                timeout=10_000,
            )
        else:
            return self.agent.get(
                f"{self.url}/v2/model/{model_id}/release/{semver}/file/{filename}/download", stream=True, timeout=10_000
            )

    def simple_upload(self, model_id: str, name: str, buffer: BytesIO):
        """Create a simple file upload.

        :param model_id: Unique model ID
        :param name: File name
        :return: JSON response object
        """
        return self.agent.post(
            f"{self.url}/v2/model/{model_id}/files/upload/simple",
            params={"name": name},
            data=buffer,
            stream=True,
            timeout=10_000,
        )

    # def start_multi_upload(): TBC

    # def finish_multi_upload(): TBC

    def delete_file(
        self,
        model_id: str,
        file_id: str,
    ):
        """Delete a specific file associated with a model.

        :param model_id: Unique model ID
        :param file_id: Unique file ID
        :return: JSON response object
        """
        return self.agent.delete(
            f"{self.url}/v2/model/{model_id}/files/{file_id}",
        ).json()

    def get_all_images(
        self,
        model_id: str,
    ):
        """Get all images.

        :param model_id: A unique model ID
        :return: JSON response object
        """
        return self.agent.get(f"{self.url}/v2/model/{model_id}/images").json()

    def get_all_schemas(
        self,
        kind: SchemaKind | None = None,
    ):
        """Get all schemas.

        :param kind: Enum to define schema kind (e.g. Model or AccessRequest), defaults to None
        :return: JSON response object
        """
        return self.agent.get(
            f"{self.url}/v2/schemas",
            params={"kind": kind},
        ).json()

    def get_schema(
        self,
        schema_id: str,
    ):
        """Retrieve a specific schema using its unique ID.

        :param schema_id: Unique schema ID
        :return: JSON response object.
        """
        return self.agent.get(
            f"{self.url}/v2/schema/{schema_id}",
        ).json()

    def post_schema(
        self,
        schema_id: str,
        name: str,
        description: str,
        kind: SchemaKind,
        json_schema: dict[str, Any],
    ):
        """Create a schema.

        :param schema_id: Unique schema ID
        :param name: Name of the schema
        :param description: Description for the schema
        :param kind: Enum to define schema kind (e.g. Model or AccessRequest)
        :param json_schema: JSON schema
        :return: JSON response object
        """
        return self.agent.post(
            f"{self.url}/v2/schemas",
            json={
                "id": schema_id,
                "name": name,
                "description": description,
                "kind": str(kind),
                "jsonSchema": json_schema,
            },
        ).json()

    def get_reviews(
        self,
        active: bool,
        model_id: str | None = None,
        version: str | None = None,
    ):
        """Get all reviews within given parameters.

        :param active: Boolean representing status of review
        :param model_id: Unique model ID, defaults to None
        :param version: Model version, defaults to None
        :return: JSON response object.
        """
        active = str(active).lower()

        return self.agent.get(
            f"{self.url}/v2/reviews",
            params={
                "active": active,
                "modelId": model_id,
                "semver": version,
            },
        ).json()

    def post_review(
        self,
        model_id: str,
        role: str,
        decision: str,
        version: str | None = None,
        comment: str | None = None,
    ):
        """Create a review for a release.

        :param model_id: A unique model ID
        :param version: A semantic version for a release
        :param role: The role of the user making the review
        :param decision: Either approve or request changes
        :param comment: A comment to go with the review
        """
        filtered_json = filter_none({"role": role, "decision": decision, "comment": comment})
        return self.agent.post(
            f"{self.url}/v2/model/{model_id}/release/{version}/review",
            json=filtered_json,
        ).json()

    def get_model_roles(
        self,
        model_id: str,
    ):
        """
        Get roles for a model.

        :param model_id: Unique model ID
        :return: JSON response object
        """
        return self.agent.get(
            f"{self.url}/v2/model/{model_id}/roles",
        ).json()

    def get_model_user_roles(
        self,
        model_id: str,
    ):
        """Get current users roles for a model.

        :param model_id: Unique model ID
        :return: JSON response object
        """
        return self.agent.get(
            f"{self.url}/v2/model/{model_id}/roles/mine",
        ).json()

    def post_team(
        self,
        team_id: str,
        name: str,
        description: str,
    ):
        """
        Create new team.

        :param team_id: Unique team ID
        :param name: Team name
        :param description: Team description
        :return: JSON response object
        """
        return self.agent.post(
            f"{self.url}/v2/teams",
            json={
                "id": team_id,
                "name": name,
                "description": description,
            },
        ).json()

    def get_all_teams(
        self,
    ):
        """
        Get all teams.

        :return: JSON response object
        """
        return self.agent.get(
            f"{self.url}/v2/teams",
        ).json()

    def get_user_teams(
        self,
    ):
        """
        Get user teams.

        :return: JSON response object
        """
        return self.agent.get(
            f"{self.url}/v2/teams/mine",
        ).json()

    def get_team(
        self,
        team_id: str,
    ):
        """Retrieve a specific team given its unique ID.

        :param team_id: Unique team ID
        :return: JSON response object
        """
        return self.agent.get(
            f"{self.url}/v2/team/{team_id}",
        ).json()

    def patch_team(
        self,
        team_id: str,
        name: str | None = None,
        description: str | None = None,
    ):
        """Update a team given its unique ID.

        :param team_id: Unique team ID
        :param name: Name of team, defaults to None
        :param description: Description of team, defaults to None
        :return: JSON response object
        """
        filtered_json = filter_none({"name": name, "description": description})

        return self.agent.patch(
            f"{self.url}/v2/team/{team_id}",
            json=filtered_json,
        ).json()

    def get_access_request(self, model_id: str, access_request_id: str):
        """Retrieve a specific access request given its unique ID.

        :param model_id: Unique model ID
        :param access_request_id: Unique access request ID
        :return: JSON response object
        """
        return self.agent.get(
            f"{self.url}/v2/model/{model_id}/access-request/{access_request_id}",
        ).json()

    def get_access_requests(
        self,
        model_id: str,
    ):
        """Retrieve all access requests given a specific model.

        :param model_id: Unique model ID
        :param access_request_id: Unique access request ID
        :return: JSON response object
        """
        return self.agent.get(
            f"{self.url}/v2/model/{model_id}/access-requests",
        ).json()

    def post_access_request(self, model_id: str, metadata: Any, schema_id: str):
        """Create an access request given a model ID.

        :param model_id: Unique model ID
        :param metadata: Metadata object, defined by access request schema
        :param schema_id: Unique schema ID
        :return: JSON response object
        """
        return self.agent.post(
            f"{self.url}/v2/model/{model_id}/access-requests",
            json={"schemaId": schema_id, "metadata": metadata},
        ).json()

    def delete_access_request(self, model_id: str, access_request_id: str):
        """Delete a specific access request associated with a model.

        :param model_id: Unique model ID
        :param access_request_id: Unique access request ID
        :return: JSON response object
        """
        return self.agent.delete(
            f"{self.url}/v2/model/{model_id}/access-request/{access_request_id}",
        ).json()

    def patch_access_request(
        self,
        model_id: str,
        access_request_id: str,
        metadata: Any,
        schema_id: str | None = None,
    ):
        """Update an access request given its unique ID.

        :param model_id: Unique model ID
        :param access_request_id: Unique access request ID
        :metadata: Metadata object, defined by access request schemas
        :return: JSON response object
        """
        filtered_json = filter_none({"schemaId": schema_id, "metadata": metadata})
        return self.agent.patch(
            f"{self.url}/v2/model/{model_id}/access-request/{access_request_id}",
            json=filtered_json,
        ).json()



from __future__ import annotations

from json import JSONDecodeError

import requests
from requests.auth import HTTPBasicAuth
from bailo.core.exceptions import BailoException, ResponseException


class Agent:
    """Base API Agent for talking with Bailo.

    Wraps each request in an exception handler that maps API errors to Python Bailo errors, among status codes less than 400.
    """

    def __request(self, method, *args, **kwargs):
        if "verify" not in kwargs:
            kwargs["verify"] = True

        res = requests.request(method, *args, **kwargs)

        # Check response for a valid range
        if res.status_code < 400:
            return res

        try:
            # Give the error message issued by bailo
            raise BailoException(res.json()["error"]["message"])
        except JSONDecodeError:
            # No response given
            raise ResponseException(f"{res.status_code} Cannot {method} to {res.request.url}")

    def get(self, *args, **kwargs):
        return self.__request("GET", *args, **kwargs)

    def post(self, *args, **kwargs):
        return self.__request("POST", *args, **kwargs)

    def patch(self, *args, **kwargs):
        return self.__request("PATCH", *args, **kwargs)

    def push(self, *args, **kwargs):
        return self.__request("PUSH", *args, **kwargs)

    def delete(self, *args, **kwargs):
        return self.__request("DELETE", *args, **kwargs)

    def put(self, *args, **kwargs):
        return self.__request("PUT", *args, **kwargs)


class PkiAgent(Agent):
    def __init__(
        self,
        cert: str,
        key: str,
        auth: str,
    ):
        """Initiate an agent for PKI authentication.

        :param cert: Path to cert file
        :param key: Path to key file
        :param auth: Path to certificate authority file
        """
        self.cert = cert
        self.key = key
        self.auth = auth

    def get(self, *args, **kwargs):
        return super().get(*args, cert=(self.cert, self.key), verify=self.auth, **kwargs)

    def post(self, *args, **kwargs):
        return super().post(*args, cert=(self.cert, self.key), verify=self.auth, **kwargs)

    def put(self, *args, **kwargs):
        return super().put(*args, cert=(self.cert, self.key), verify=self.auth, **kwargs)

    def patch(self, *args, **kwargs):
        return super().patch(*args, cert=(self.cert, self.key), verify=self.auth, **kwargs)

    def delete(self, *args, **kwargs):
        return super().delete(*args, cert=(self.cert, self.key), verify=self.auth, **kwargs)


class TokenAgent(Agent):
    def __init__(
        self,
        access_key: str,
        secret_key: str,
    ):
        """Initiate an agent for API token authentication.

        :param access_key: Access key
        :param secret_key: Secret key
        """
        self.access_key = access_key
        self.secret_key = secret_key
        self.basic = HTTPBasicAuth(access_key, secret_key)

    def get(self, *args, **kwargs):
        return super().get(*args, auth=self.basic, **kwargs)

    def post(self, *args, **kwargs):
        return super().post(*args, auth=self.basic, **kwargs)

    def put(self, *args, **kwargs):
        return super().put(*args, auth=self.basic, **kwargs)

    def patch(self, *args, **kwargs):
        return super().patch(*args, auth=self.basic, **kwargs)

    def delete(self, *args, **kwargs):
        return super().delete(*args, auth=self.basic, **kwargs)



from __future__ import annotations

from typing import Any

from bailo.core.client import Client
from bailo.core.enums import SchemaKind


class Schema:
    """Represent a schema within Bailo.

    :param client: A client object used to interact with Bailo
    :param schema_id: A unique schema ID
    :param name: Name of schema
    :param description: Description of the schema
    :param kind: Kind of schema, using SchemaKind enum (e.g Model or AccessRequest)
    :param json_schema: Schema JSON
    """

    def __init__(
        self,
        client: Client,
        schema_id: str,
        name: str,
        description: str,
        kind: SchemaKind,
        json_schema: dict[str, Any],
    ) -> None:
        self.client = client
        self.schema_id = schema_id
        self.name = name
        self.description = description
        self.kind = kind
        self.json_schema = json_schema

    @classmethod
    def create(
        cls,
        client: Client,
        schema_id: str,
        name: str,
        description: str,
        kind: SchemaKind,
        json_schema: dict[str, Any],
    ) -> Schema:
        """Build a schema from Bailo and uploads it.

        :param client: A client object used to interact with Bailo
        :param schema_id: A unique schema ID
        :param name: Name of schema
        :param description: Description of schema
        :param kind: Kind of schema, using SchemaKind enum (e.g Model or AccessRequest)
        :param json_schema: Schema JSON
        :return: Schema object
        """
        schema = cls(
            client=client,
            schema_id=schema_id,
            name=name,
            description=description,
            kind=kind,
            json_schema=json_schema,
        )
        res = client.post_schema(
            schema_id=schema_id, name=name, description=description, kind=kind, json_schema=json_schema
        )
        schema.__unpack(res["schema"])

        return schema

    @classmethod
    def from_id(cls, client: Client, schema_id: str) -> Schema:
        """Return an existing schema from Bailo.

        :param client: A client object used to interact with Bailo
        :param schema_id: A unique schema ID
        :return: Schema object
        """
        schema = cls(
            client=client,
            schema_id=schema_id,
            name="temp",
            description="temp",
            kind=SchemaKind.MODEL,
            json_schema={"temp": "temp"},
        )
        res = client.get_schema(schema_id=schema_id)
        schema.__unpack(res["schema"])

        return schema

    def __unpack(self, res) -> None:
        self.schema_id = res["id"]
        self.name = res["name"]
        self.description = res["description"]
        kind = res["kind"]
        self.json_schema = res["jsonSchema"]

        if kind == "model":
            self.kind = SchemaKind.MODEL
        if kind == "accessRequest":
            self.kind = SchemaKind.ACCESS_REQUEST



from __future__ import annotations

from typing import Any


def filter_none(json: dict[str, Any]) -> dict[str, Any]:
    """Remove None attributes from a dictionary.

    :param json: Dictionary to filter
    :return: Dictionary with removed Nonetypes
    """
    res = {k: v for k, v in json.items() if v is not None}
    return res


class NestedDict(dict):
    def __getitem__(self, keytuple):
        # if key is not a tuple then access as normal
        if not isinstance(keytuple, tuple):
            return super(NestedDict, self).__getitem__(keytuple)
        d = self
        for key in keytuple:
            d = d[key]
        return d

    def __setitem__(self, keytuple, item):
        # if key is not a tuple then access as normal
        if not isinstance(keytuple, tuple):
            return super(NestedDict, self).__setitem__(keytuple, item)
        d = self
        for index, key in enumerate(keytuple):
            if index != len(keytuple) - 1:
                d = d.setdefault(key, {})
            else:
                d[key] = item



"""
Bailo Python Client
===================

Bailo is a ecosystem for managing the lifecycle of managing machine learning models. This package provides support for interacting with models within Bailo.
"""
from __future__ import annotations

__version__ = "2.1.1"

from bailo.core.agent import Agent, PkiAgent, TokenAgent
from bailo.core.client import Client
from bailo.core.enums import ModelVisibility, Role, SchemaKind
from bailo.helper.access_request import AccessRequest
from bailo.helper.model import Model, Experiment
from bailo.helper.release import Release
from bailo.helper.schema import Schema



"""An object orientated class structure for Bailo's API.

This module allows interaction of Models, Releases, Access Requests and Schemas.


>>> from bailo import Model, Client
>>>
>>> client = Client("https://bailo.com") # URL of Bailo

To import a resource from Bailo:

>>> my_model = Model.from_id("yolov4")

To create a resource that will be automatically reflected in Bailo:

>>> my_model = Model.create(
    client=client,
    name="YoloV4",
    description="You only look once!",
    team_id="Uncategorised"
)
"""
